@@ -31,8 +31,10 @@ static char rcsid[] = "$Header";
 #ifndef MOTIF11
 #include <Xm/DragDrop.h>
 #endif
+#include <Xm/Command.h>
 #include <Xm/List.h>
 #include <Xm/Text.h>
+#include <Xm/RowColumn.h>
 		
 /*--------------------------------------------------------------
  *
@@ -83,7 +85,7 @@ Tm_AnyWidgetCmd(clientData, interp, argc, argv)
 	if (argc < 3) {
 	    sprintf(interp->result, "wrong # args: should be \"%.50s callActionProc \
 action", argv[0]);
-	    goto error;
+	    return TCL_ERROR;
 	}
 
 	if (Tm_MakeXEvent(wPtr->widget, interp, &xev, argc-3, argv+3) == TCL_ERROR)
@@ -106,14 +108,47 @@ action", argv[0]);
 	return TCL_OK;
     } else
 
+    if ((c == 'd') && (strcmp(argv[1], "defineCursor") == 0)) {
+	XrmValue from, to;
+	Cursor cursor;
+
+	if (argc < 3) {
+	    sprintf(interp->result, "wrong # args: should be \"%.50s defineCursor \
+value", argv[0]);
+	    return TCL_ERROR;
+	}
+
+	if (!XtIsRealized(w)) {
+            sprintf(interp->result, "widget must be realized: \"%.50s defineCursor \
+value", argv[0]);
+            return TCL_ERROR;
+        }
+
+	/* special case: "" restores default */
+	if (argv[2][0] == '\0') {
+	    XDefineCursor(XtDisplay(w), XtWindow(w), None);
+	    return TCL_OK;
+	}
+
+	from.addr = argv[2];
+	from.size = sizeof(String);
+	to.size = sizeof(Cursor *);
+	to.addr = (caddr_t) &cursor;
+
+	if (XtConvertAndStore(w, XmRString, &from,
+				XmRCursor, &to)) {
+	    XDefineCursor(XtDisplay(w), XtWindow(w),* (Cursor *) to.addr);
+	} else {
+	    sprintf(interp->result, "wrong cursor to \"%.50s defineCursor \
+value", argv[0]);
+	    return TCL_ERROR;
+	}
+    } else
+
     if ((c == 'd') && (strcmp(argv[1], "destroyWidget") == 0)) {
 	/* destroy - most stuff done in DestroyWidgetHandler */
 
-        Tcl_DeleteCommand(interp, wPtr->pathName);
-        XtFree(wPtr->pathName);
-
 	XtDestroyWidget(wPtr->widget);
-        ckfree(wPtr);
     } else
 
 #ifndef MOTIF11
@@ -149,11 +184,20 @@ action", argv[0]);
     } else
 #endif
 
+    if ((c == 'g') && (strcmp(argv[1], "getAppResources") == 0)) {
+	if (argc < 3) {
+	    sprintf(interp->result, "wrong # args: should be \"%.50s getAppResources \
+resourceList\"", argv[0]);
+	    return TCL_ERROR;
+	}
+	return(Tm_GetAppResources(interp, w, argv[2]));
+    } else
+
     if ((c == 'g') && (strcmp(argv[1], "getGC") == 0)) {
 	if (argc < 4) {
 	    sprintf(interp->result, "wrong # args: should be \"%.50s setValues \
 -option value...", argv[0]);
-	    goto error;
+	    return TCL_ERROR;
 	}
 	interp->result =
 	    Tm_GetGC(wPtr->pathName, interp, (Widget) wPtr->widget, 
@@ -164,11 +208,53 @@ action", argv[0]);
 	if (argc < 4) {
 	    sprintf(interp->result, "wrong # args: should be \"%.50s setValues \
 -option value...", argv[0]);
-	    goto error;
+	    return TCL_ERROR;
 	}
+        Tm_InitFreeResourceList((argc - 2)/2);
 	if (Tm_GetValues(wPtr->pathName, interp, (Widget) wPtr->widget, 
 			XtClass(wPtr->widget), argv+2, argc-2) == TCL_ERROR)
 	    return TCL_ERROR;
+	Tm_FreeResourceList();
+    } else
+
+    if ((c == 'i') && (strcmp(argv[1], "isComposite") == 0)) {
+	if (XtIsComposite(wPtr->widget)) {
+	    interp->result = "true";
+	} else {
+	    interp->result = "false";
+	}
+    } else
+
+    if ((c == 'i') && (strcmp(argv[1], "isConstraint") == 0)) {
+	if (XtIsConstraint(wPtr->widget)) {
+	    interp->result = "true";
+	} else {
+	    interp->result = "false";
+	}
+    } else
+
+    if ((c == 'i') && (strcmp(argv[1], "isManaged") == 0)) {
+	if (XtIsManaged(wPtr->widget)) {
+	    interp->result = "true";
+	} else {
+	    interp->result = "false";
+	}
+    } else
+
+    if ((c == 'i') && (strcmp(argv[1], "isSensitive") == 0)) {
+	if (XtIsSensitive(wPtr->widget)) {
+	    interp->result = "true";
+	} else {
+	    interp->result = "false";
+	}
+    } else
+
+    if ((c == 'i') && (strcmp(argv[1], "isShell") == 0)) {
+	if (XtIsShell(wPtr->widget)) {
+	    interp->result = "true";
+	} else {
+	    interp->result = "false";
+	}
     } else
 
     if ((c == 'm') && (strcmp(argv[1], "manageChild") == 0)) {
@@ -192,7 +278,7 @@ action", argv[0]);
 	if (argc != 3) {
 	    sprintf(interp->result, "wrong # args: should be \"%.50s processTraversal \
 direction", argv[0]);
-	    goto error;
+	    return TCL_ERROR;
 	}
 
 	if (strcmp(argv[2], "current") == 0) {
@@ -232,7 +318,7 @@ direction", argv[0]);
 	XmProcessTraversal(w, direction);
 	return TCL_OK;
     } else
-#endif
+#endif /* MOTIF11 */
 
     if ((c == 'r') && (strcmp(argv[1], "realizeWidget") == 0)) {
 	XtRealizeWidget(wPtr->widget);
@@ -247,7 +333,7 @@ direction", argv[0]);
 	if (argc < 3) {
 	    sprintf(interp->result, "wrong # args: should be \"%.50s setSensitive \
 -option value", argv[0]);
-	    goto error;
+	    return TCL_ERROR;
 	}
 	if (Tcl_GetBoolean(interp, argv[2], &bool_val) == TCL_ERROR) {
 	    return TCL_ERROR;
@@ -259,8 +345,10 @@ direction", argv[0]);
 	if (argc < 4) {
 	    sprintf(interp->result, "wrong # args: should be \"%.50s setValues \
 -option value...", argv[0]);
-	    goto error;
+	    return TCL_ERROR;
 	}
+	Tm_InitFreeResourceList((argc - 2)/2);
+
 	if ( ! XtIsShell(wPtr->widget)) {
 	    /* this case to prevent Xt errors on XtParent of "." */
 	    if (Tm_SetValues(wPtr->pathName, interp, wPtr->widget,
@@ -276,6 +364,7 @@ direction", argv[0]);
 		return TCL_ERROR;
 	}
 	XtSetValues(wPtr->widget, args, num_args);
+	Tm_FreeResourceList();
     } else
     if ((c == 'u') && (strcmp(argv[1], "unmanageChild") == 0)) {
 	/* unmanage child */
@@ -290,10 +379,20 @@ direction", argv[0]);
 	Tm_ClientData *client_data;
 
 	client_data = (Tm_ClientData *) XtMalloc(sizeof(Tm_ClientData));
+/*
 	client_data->callback_func = XtNewString(argv[2]);
+*/
+	if (argc == 3) {
+	    /* for backward compatability with eg argv[2] = {do it} */
+	    client_data->callback_func = XtNewString(argv[2]);
+	} else {
+	    client_data->callback_func = Tcl_Merge(argc - 2, argv + 2);
+	}
 	client_data->widget_info = wPtr;
 	XtAddCallback(wPtr->widget, argv[1], Tm_WidgetCallbackHandler,
 		(XtPointer) client_data);
+	XtAddCallback(wPtr->widget, XmNdestroyCallback, 
+		Tm_DestroyReclaimHandler, (XtPointer) client_data);
     } else {
 	goto error;
     }
@@ -471,6 +570,10 @@ Tm_ListWidgetCmd (clientData, interp, argc, argv)
 	/* don't free this sucker! DestroyWidget will look after it!!!
 	    XmStringFree(*(XmString *) toItem.addr);
 	 */
+	/* using our own converters, we free it now 
+	 */
+	XmStringFree(*(XmString *) toItem.addr);
+
 /*
 	xmstr = XmStringCreateLocalized(argv[2]);
 	XmListAddItem(w, xmstr, position);
@@ -505,6 +608,7 @@ Tm_ListWidgetCmd (clientData, interp, argc, argv)
 	/* don't free this sucker! DestroyWidget will look after it!!!
 	    XmStringFree(*(XmString *) toItem.addr);
 	 */
+	XmStringFree(*(XmString *) toItem.addr);
     } else
 
 #ifndef MOTIF11
@@ -532,6 +636,9 @@ Tm_ListWidgetCmd (clientData, interp, argc, argv)
 	/* don't free this sucker! DestroyWidget will look after it!!!
 	    XmStringFree(*(XmString *) toItem.addr);
 	 */
+	/* using our converter, do so */
+	XmStringFree(*(XmString *) toItem.addr);
+	
     } else
 
     if ((c == 'd') && (strcmp(argv[1], "deletePosition") == 0)) {
@@ -572,6 +679,9 @@ Tm_ListWidgetCmd (clientData, interp, argc, argv)
 	/* don't free this sucker! DestroyWidget will look after it!!!
 	    XmStringFree(*(XmString *) toItem.addr);
 	 */
+	/* using our converter, free it */
+	XmStringFree(*(XmString *) toItem.addr);
+
     } else
 
     if ((c == 'd') && (strcmp(argv[1], "deselectPosition") == 0)) {
@@ -608,6 +718,7 @@ Tm_ListWidgetCmd (clientData, interp, argc, argv)
 	} else {
 	    sprintf(interp->result, "false");
 	}
+	XmStringFree(*(XmString *) toItem.addr);
     } else
 
 #ifndef MOTIF11
@@ -628,6 +739,7 @@ Tm_ListWidgetCmd (clientData, interp, argc, argv)
 
 	position = XmListItemPos(w, *(XmString *)toItem.addr);
 	sprintf(interp->result, "%d", position);
+	XmStringFree(*(XmString *)toItem.addr);
     } else
 #endif
 
@@ -673,6 +785,7 @@ Tm_ListWidgetCmd (clientData, interp, argc, argv)
         }
 
 	XmListSelectItem(w, *(XmString *) toItem.addr, call_callback);
+	XmStringFree(*(XmString *) toItem.addr);
     } else
 
     if ((c == 's') && (strcmp(argv[1], "selectPosition") == 0)) {
@@ -709,6 +822,8 @@ Tm_ListWidgetCmd (clientData, interp, argc, argv)
         XtConvertAndStore(w, XtRString, &from, XmRXmString, &toItem);
 
 	XmListSetBottomItem(w, *(XmString *) toItem.addr);
+	XmStringFree(*(XmString *) toItem.addr);
+
     } else
 
     if ((c == 's') && (strcmp(argv[1], "setBottomPosition") == 0)) {
@@ -742,6 +857,8 @@ Tm_ListWidgetCmd (clientData, interp, argc, argv)
         XtConvertAndStore(w, XtRString, &from, XmRXmString, &toItem);
 
 	XmListSetItem(w, *(XmString *) toItem.addr);
+	XmStringFree(*(XmString *) toItem.addr);
+
     } else
 
     if ((c == 's') && (strcmp(argv[1], "setPosition") == 0)) {
@@ -860,15 +977,15 @@ permissions tcl_code", argv[0]);
 	/* determine modes */
 	if (ch == 'r') {
 	    inputId = XtAppAddInput(appContext, filenum, (XtPointer) XtInputReadMask,
-			Tm_InputHandler, input);
+			Tm_InputHandler, (XtPointer) input);
 	} else
 	if (ch == 'w') {
 	    inputId = XtAppAddInput(appContext, filenum, (XtPointer) XtInputWriteMask,
-			Tm_InputHandler, input);
+			Tm_InputHandler, (XtPointer) input);
 	} else
 	if (ch == 'x') {
 	    inputId = XtAppAddInput(appContext, filenum, (XtPointer) XtInputExceptMask,
-			Tm_InputHandler, input);
+			Tm_InputHandler, (XtPointer) input);
 	}
 	
 	/* simple type check: prefix Id with "inputId-" */
@@ -897,21 +1014,12 @@ permissions tcl_code", argv[0]);
 	timer->interp = interp;
 	timer->command = XtNewString(argv[3]);
 	timerId = XtAppAddTimeOut(appContext, interval,
-                        Tm_TimerHandler, timer);
+                        Tm_TimerHandler, (XtPointer) timer);
 
 	sprintf(interp->result, "timerId-%lu", timerId);
 	return TCL_OK;
     } else
 
-    if ((c == 'g') && (strcmp(argv[1], "getAppResources") == 0)) {
-	if (argc < 3) {
-	    sprintf(interp->result, "wrong # args: should be \"%.50s getAppResources \
-resourceList\"", argv[0]);
-	    return TCL_ERROR;
-	}
-	return(Tm_GetAppResources(interp, w, argv[2]));
-    } else
-
     if ((c == 'p') && (strcmp(argv[1], "processEvent") == 0)) {
 	XtAppProcessEvent(appContext, XtIMAll);
     } else
@@ -956,7 +1064,7 @@ timerID\"", argv[0]);
 	XtAppMainLoop(XtWidgetToApplicationContext(wPtr->widget));
 
     } else {
-        return Tm_AnyWidgetCmd(clientData, interp, argc, argv);
+        return Tm_ShellWidgetCmd(clientData, interp, argc, argv);
     }
     return TCL_OK;
 }
@@ -1020,7 +1128,7 @@ Tm_DrawnWidgetCmd (clientData, interp, argc, argv)
 /*	%p may be broken on the Sun? So fit it into an XtArgVal
 	sscanf(argv[2], "%p", &p);
 */
-	if (sscanf(argv[2], "gc-%ld", &p) < 1) {;
+	if (sscanf(argv[2], "gc-%lu", &p) < 1) {;
 	    sprintf(interp->result, "drawImageString: illegal gc \"%s\"",
 			argv[2]);
 	    return TCL_ERROR;
@@ -1035,6 +1143,72 @@ Tm_DrawnWidgetCmd (clientData, interp, argc, argv)
 	win = XtWindow(w);
 
 	XDrawImageString(display, win, gc, x, y, string, length); 
+    } else if ((c == 'd') && (strcmp(argv[1], "drawArc") == 0)) {
+
+	long p;
+	GC gc;
+	int x, y;
+	unsigned int width, height;
+	int angle1, angle2;
+	Window win;
+	Display *display;
+	int length;
+	char *string;
+
+        if (argc != 9) {
+    	    sprintf(interp->result,
+		    "wrong # args: should be \"%.50s drawArc gc x y width height angle1 angle2\"",
+		    argv[0]);
+	    return TCL_ERROR;
+        }
+	if (sscanf(argv[2], "gc-%lu", &p) < 1) {;
+	    sprintf(interp->result, "drawArc: illegal gc \"%s\"",
+			argv[2]);
+	    return TCL_ERROR;
+	}
+	gc = (GC) p;
+	Tcl_GetInt(interp, argv[3], &x);
+	Tcl_GetInt(interp, argv[4], &y);
+	Tcl_GetInt(interp, argv[5], &width);
+	Tcl_GetInt(interp, argv[6], &height);
+	Tcl_GetInt(interp, argv[7], &angle1);
+	Tcl_GetInt(interp, argv[8], &angle2);
+
+	display = XtDisplay(w);
+	win = XtWindow(w);
+
+	XDrawArc(display, win, gc, x, y, width, height, angle1, angle2); 
+    } else if ((c == 'd') && (strcmp(argv[1], "drawLine") == 0)) {
+
+	long p;
+	GC gc;
+	int x1, y1, x2, y2;
+	Window win;
+	Display *display;
+	int length;
+	char *string;
+
+        if (argc != 7) {
+    	    sprintf(interp->result,
+		    "wrong # args: should be \"%.50s drawLine gc x1 y1 x2 y2\"",
+		    argv[0]);
+	    return TCL_ERROR;
+        }
+	if (sscanf(argv[2], "gc-%lu", &p) < 1) {;
+	    sprintf(interp->result, "drawLine: illegal gc \"%s\"",
+			argv[2]);
+	    return TCL_ERROR;
+	}
+	gc = (GC) p;
+	Tcl_GetInt(interp, argv[3], &x1);
+	Tcl_GetInt(interp, argv[4], &y1);
+	Tcl_GetInt(interp, argv[5], &x2);
+	Tcl_GetInt(interp, argv[6], &y2);
+
+	display = XtDisplay(w);
+	win = XtWindow(w);
+
+	XDrawLine(display, win, gc, x1, y1, x2, y2); 
     } else {
         return Tm_AnyWidgetCmd(clientData, interp, argc, argv);
     }
@@ -1093,6 +1267,8 @@ Tm_TextWidgetCmd (clientData, interp, argc, argv)
 
 	XtAddCallback(wPtr->widget, XmNlosingFocusCallback,
 			Tm_TextVerifyCallbackHandler, (XtPointer) client_data);
+        XtAddCallback(wPtr->widget, XmNdestroyCallback, 
+                Tm_DestroyReclaimHandler, (XtPointer) client_data);
     } else
 
     if (strcmp(argv[1], "modifyVerifyCallback") == 0) {
@@ -1110,6 +1286,8 @@ Tm_TextWidgetCmd (clientData, interp, argc, argv)
 
 	XtAddCallback(wPtr->widget, XmNmodifyVerifyCallback,
 			Tm_TextVerifyCallbackHandler, (XtPointer) client_data);
+        XtAddCallback(wPtr->widget, XmNdestroyCallback, 
+                Tm_DestroyReclaimHandler, (XtPointer) client_data);
     } else
 
     if (strcmp(argv[1], "motionVerifyCallback") == 0) {
@@ -1127,6 +1305,8 @@ Tm_TextWidgetCmd (clientData, interp, argc, argv)
 
 	XtAddCallback(wPtr->widget, XmNmotionVerifyCallback,
 			Tm_TextVerifyCallbackHandler, (XtPointer) client_data);
+        XtAddCallback(wPtr->widget, XmNdestroyCallback, 
+                Tm_DestroyReclaimHandler, (XtPointer) client_data);
     } else
 
     if (strcmp(argv[1], "clearSelection") == 0) {
@@ -1596,3 +1776,94 @@ Tm_TextWidgetCmd (clientData, interp, argc, argv)
     }
     return TCL_OK;
 }
+
+/*
+ *--------------------------------------------------------------
+ *
+ * Tm_PopupMenuWidgetCmd --
+ *
+ *	handles the methods that may be requested of Popup menus.
+ *
+ * Results:
+ *
+ * Side effects:
+ *
+ *--------------------------------------------------------------
+ */
+
+int
+Tm_PopupMenuWidgetCmd (clientData, interp, argc, argv)
+    ClientData clientData;
+    Tcl_Interp *interp;
+    int argc;
+    char **argv;
+{
+    Tm_Widget *wPtr = (Tm_Widget *) clientData;
+    Widget w = wPtr->widget;
+    XButtonPressedEvent *event;
+
+    if (strcmp(argv[1], "menuPosition") == 0) {
+        if (argc != 3) {
+            sprintf(interp->result,
+                "wrong # args: should be \"%.50s menuPosition event\"",
+                argv[0]);
+            return TCL_ERROR;
+        }
+	if (sscanf(argv[2], "event-%lu", &event) < 1) {
+            sprintf(interp->result,
+                "wrong event arg: should be \"%.50s menuPosition event\"",
+                argv[0]);
+            return TCL_ERROR;
+	}
+	XmMenuPosition(w, event);
+	return TCL_OK;
+    } else {
+        return Tm_AnyWidgetCmd(clientData, interp, argc, argv);
+    }
+    return TCL_OK;
+}
+
+/*
+ *--------------------------------------------------------------
+ *
+ * Tm_ShellWidgetCmd --
+ *
+ *	handles the methods that may be requested of Shells
+ *
+ * Results:
+ *
+ * Side effects:
+ *
+ *--------------------------------------------------------------
+ */
+
+int
+Tm_ShellWidgetCmd (clientData, interp, argc, argv)
+    ClientData clientData;
+    Tcl_Interp *interp;
+    int argc;
+    char **argv;
+{
+    Tm_Widget *wPtr = (Tm_Widget *) clientData;
+    Widget w = wPtr->widget;
+
+    if (strcmp(argv[1], "popup") == 0) {
+/*	should we allow grab type to be set?
+        if (argc != 3) {
+            sprintf(interp->result,
+                "wrong # args: should be \"%.50s menuPosition event\"",
+                argv[0]);
+            return TCL_ERROR;
+        }
+*/
+	XtPopup(w, XtGrabNone);
+	return TCL_OK;
+    } else
+    if (strcmp(argv[1], "popup") == 0) {
+	XtPopdown(w);
+	return TCL_OK;
+    } else {
+        return Tm_AnyWidgetCmd(clientData, interp, argc, argv);
+    }
+    return TCL_OK;
+}
