@@ -37,6 +37,7 @@ static char rcsid[] = "$Header: /usrs/tm/RCS/tmPushButton.c,v 1.2 1993/07/14 20:
 #include <Xm/BulletinB.h>
 #include <Xm/CascadeB.h>
 #include <Xm/Command.h>
+#include <Xm/DialogS.h>
 #include <Xm/DrawingA.h>
 #include <Xm/DrawnB.h>
 #include <Xm/FileSB.h>
@@ -59,6 +60,8 @@ static char rcsid[] = "$Header: /usrs/tm/RCS/tmPushButton.c,v 1.2 1993/07/14 20:
 #include <Xm/ToggleB.h>
 #endif  /* MOTIF11 */
 
+#include <X11/Shell.h>
+
 /*
  * this is the list of resources that require the widget rather
  * than its parent for proper conversions from String.
@@ -87,6 +90,7 @@ static char *dont_use_parent_for_resources[] = {
 	"sashWidth",
 	"sashHeight",
 	"labelPixmap",
+	"value",	/* need to know if it came from Text, so we can gc it */
 /*	"labelType",	 this needn't be here, except size calcs go wrong */
 };
 
@@ -228,74 +232,80 @@ Tm_AnyCmd(clientData, interp, argc, argv)
     if (parent == NULL)
 	return TCL_ERROR;
 
-    if (strcmp(command, ArrowButton) == 0) {
+    if (strcmp(command, xmArrowButton) == 0) {
 	class = xmArrowButtonWidgetClass;
     } else
-    if (strcmp(command, BulletinBoard) == 0) {
+    if (strcmp(command, xmBulletinBoard) == 0) {
 	class = xmBulletinBoardWidgetClass;
     } else
-    if (strcmp(command, CascadeButton) == 0) {
+    if (strcmp(command, xmCascadeButton) == 0) {
 	class = xmCascadeButtonWidgetClass;
     } else
-    if (strcmp(command, Command) == 0) {
+    if (strcmp(command, xmCommand) == 0) {
 	class = xmCommandWidgetClass;
     } else
-    if (strcmp(command, DrawingArea) == 0) {
+    if (strcmp(command, xmDialogShell) == 0) {
+	class = xmDialogShellWidgetClass;
+    } else
+    if (strcmp(command, xmDrawingArea) == 0) {
 	class = xmDrawingAreaWidgetClass;
     } else
-    if (strcmp(command, DrawnButton) == 0) {
+    if (strcmp(command, xmDrawnButton) == 0) {
 	class = xmDrawnButtonWidgetClass;
     } else
-    if (strcmp(command, FileSelectionBox) == 0) {
+    if (strcmp(command, xmFileSelectionBox) == 0) {
 	class = xmFileSelectionBoxWidgetClass;
     } else
-    if (strcmp(command, Form) == 0) {
+    if (strcmp(command, xmForm) == 0) {
 	class = xmFormWidgetClass;
     } else
-    if (strcmp(command, Frame) == 0) {
+    if (strcmp(command, xmFrame) == 0) {
 	class = xmFrameWidgetClass;
     } else
-    if (strcmp(command, List) == 0) {
+    if (strcmp(command, xmList) == 0) {
 	class = xmListWidgetClass;
     } else
-    if (strcmp(command, Label) == 0) {
+    if (strcmp(command, xmLabel) == 0) {
 	class = xmLabelWidgetClass;
     } else
-    if (strcmp(command, MainWindow) == 0) {
+    if (strcmp(command, xmMainWindow) == 0) {
 	class = xmMainWindowWidgetClass;
     } else
-    if (strcmp(command, MessageBox) == 0) {
+    if (strcmp(command, xmMessageBox) == 0) {
 	class = xmMessageBoxWidgetClass;
     } else
-    if (strcmp(command, PanedWindow) == 0) {
+    if (strcmp(command, xmPanedWindow) == 0) {
 	class = xmPanedWindowWidgetClass;
     } else
-    if (strcmp(command, PushButton) == 0) {
+    if (strcmp(command, xmPushButton) == 0) {
 	class = xmPushButtonWidgetClass;
     } else
-    if (strcmp(command, Scale) == 0) {
+    if (strcmp(command, xmScale) == 0) {
 	class = xmScaleWidgetClass;
     } else
-    if (strcmp(command, ScrollBar) == 0) {
+    if (strcmp(command, xmScrollBar) == 0) {
 	class = xmScrollBarWidgetClass;
     } else
-    if (strcmp(command, ScrolledWindow) == 0) {
+    if (strcmp(command, xmScrolledWindow) == 0) {
 	class = xmScrolledWindowWidgetClass;
     } else
-    if (strcmp(command, SelectionBox) == 0) {
+    if (strcmp(command, xmSelectionBox) == 0) {
 	class = xmSelectionBoxWidgetClass;
     } else
-    if (strcmp(command, Separator) == 0) {
+    if (strcmp(command, xmSeparator) == 0) {
 	class = xmSeparatorWidgetClass;
     } else
-    if (strcmp(command, Text) == 0) {
+    if (strcmp(command, xmText) == 0) {
 	class = xmTextWidgetClass;
     } else
-    if (strcmp(command, TextField) == 0) {
+    if (strcmp(command, xmTextField) == 0) {
 	class = xmTextFieldWidgetClass;
     } else
-    if (strcmp(command, ToggleButton) == 0) {
+    if (strcmp(command, xmToggleButton) == 0) {
 	class = xmToggleButtonWidgetClass;
+    } else
+    if (strcmp(command, xmTopLevelShell) == 0) {
+	class = topLevelShellWidgetClass;
     } else {
 	return TCL_ERROR;
     }
@@ -311,12 +321,19 @@ Tm_AnyCmd(clientData, interp, argc, argv)
 
     DivideArgs(argv + 2, argc - 2, parent_argv, &num_parent_argc,
 			child_argv, &num_child_argc);
+
+    /* set up for garbage collect of some resource values */
+    Tm_InitFreeResourceList((argc - 2)/2);
+
     Tm_SetValues(path, interp, parent, parent, class,
 				parent_argv, num_parent_argc, args, &num_args);
     new = XtCreateWidget (Tm_NameFromPath(path), class, parent, args, num_args);
     Tm_SetValues(path, interp, new, parent, class,
 				child_argv, num_child_argc, args, &num_args);
     XtSetValues(new, args, num_args);
+
+    /* collect garbage */
+    Tm_FreeResourceList();
     
     if (managed) {
         XtManageChild (new);
@@ -384,7 +401,7 @@ CreateChild(interp, path, w, name)
     wPtr -> interp = interp;
     wPtr -> widget = w;
     wPtr -> pathName = subobject_path;
-    wPtr -> parent = path;
+    wPtr -> parent = XtNewString(path);
     wPtr -> dropProc = NULL;
     wPtr -> displayInfo = ((Tm_Widget *) cmdInfo.clientData)->displayInfo;
 
@@ -562,7 +579,7 @@ Tm_HasChildrenCmd(clientData, interp, argc, argv)
 
     new = ((Tm_Widget *) cmdInfo.clientData)->widget;
 
-    if (strcmp(command, SelectionBox) == 0) {
+    if (strcmp(command, xmSelectionBox) == 0) {
 	SelectionBoxCreateChild(interp, path, new, XmDIALOG_APPLY_BUTTON, "Apply");
 	SelectionBoxCreateChild(interp, path, new, XmDIALOG_CANCEL_BUTTON, "Cancel");
 	SelectionBoxCreateChild(interp, path, new, XmDIALOG_HELP_BUTTON, "Help");
@@ -576,7 +593,7 @@ Tm_HasChildrenCmd(clientData, interp, argc, argv)
 	return TCL_OK;
     }
 
-    if (strcmp(command, MessageBox) == 0) {
+    if (strcmp(command, xmMessageBox) == 0) {
 	MessageBoxCreateChild(interp, path, new, XmDIALOG_CANCEL_BUTTON, "Cancel");
 	MessageBoxCreateChild(interp, path, new, XmDIALOG_HELP_BUTTON, "Help");
 	MessageBoxCreateChild(interp, path, new, XmDIALOG_MESSAGE_LABEL, "Message");
@@ -587,7 +604,7 @@ Tm_HasChildrenCmd(clientData, interp, argc, argv)
 	return TCL_OK;
     }
 
-    if (strcmp(command, FileSelectionBox) == 0) {
+    if (strcmp(command, xmFileSelectionBox) == 0) {
 	FileSelectionBoxCreateChild(interp, path, new, XmDIALOG_APPLY_BUTTON, "Apply");
 	FileSelectionBoxCreateChild(interp, path, new, XmDIALOG_CANCEL_BUTTON, "Cancel");
 	FileSelectionBoxCreateChild(interp, path, new, XmDIALOG_DIR_LIST, "DirList");
@@ -605,7 +622,7 @@ Tm_HasChildrenCmd(clientData, interp, argc, argv)
 	return TCL_OK;
     }
 
-    if (strcmp(command, ScrolledWindow) == 0) {
+    if (strcmp(command, xmScrolledWindow) == 0) {
 	ScrolledWindowCreateChild(interp, path, new, XmNhorizontalScrollBar,
 				"HorScrollBar");
 	ScrolledWindowCreateChild(interp, path, new, XmNverticalScrollBar,
@@ -681,7 +698,7 @@ Tm_DialogCmd(clientData, interp, argc, argv)
     DivideArgs(argv + 2, argc - 2, parent_argv, &num_parent_argc,
                         child_argv, &num_child_argc);
 
-    if (strcmp(command, BulletinBoardDialog) == 0) {
+    if (strcmp(command, xmBulletinBoardDialog) == 0) {
 	class = xmBulletinBoardWidgetClass;
         XtInitializeWidgetClass(class);
         Tm_SetValues(path, interp, parent, parent, class,
@@ -689,7 +706,7 @@ Tm_DialogCmd(clientData, interp, argc, argv)
 	new = XmCreateBulletinBoardDialog (parent,
 			Tm_NameFromPath(path), args, num_args);
     } else
-    if (strcmp(command, FileSelectionDialog) == 0) {
+    if (strcmp(command, xmFileSelectionDialog) == 0) {
 	class = xmFileSelectionBoxWidgetClass;
         XtInitializeWidgetClass(class);
         Tm_SetValues(path, interp, parent, parent, class,
@@ -697,7 +714,7 @@ Tm_DialogCmd(clientData, interp, argc, argv)
 	new = XmCreateFileSelectionDialog (parent,
 			Tm_NameFromPath(path), args, num_args);
     } else
-    if (strcmp(command, FormDialog) == 0) {
+    if (strcmp(command, xmFormDialog) == 0) {
 	class = xmFormWidgetClass;
         XtInitializeWidgetClass(class);
         Tm_SetValues(path, interp, parent, parent, class,
@@ -705,7 +722,7 @@ Tm_DialogCmd(clientData, interp, argc, argv)
 	new = XmCreateFormDialog (parent,
 			Tm_NameFromPath(path), args, num_args);
     } else
-    if (strcmp(command, ErrorDialog) == 0) {
+    if (strcmp(command, xmErrorDialog) == 0) {
 	class = xmMessageBoxWidgetClass;
         XtInitializeWidgetClass(class);
         Tm_SetValues(path, interp, parent, parent, class,
@@ -713,7 +730,7 @@ Tm_DialogCmd(clientData, interp, argc, argv)
 	new = XmCreateErrorDialog (parent, 
 			Tm_NameFromPath(path), args, num_args);
     } else 
-    if (strcmp(command, InformationDialog) == 0) {
+    if (strcmp(command, xmInformationDialog) == 0) {
 	class = xmMessageBoxWidgetClass;
         XtInitializeWidgetClass(class);
         Tm_SetValues(path, interp, parent, parent, class,
@@ -721,7 +738,7 @@ Tm_DialogCmd(clientData, interp, argc, argv)
 	new = XmCreateInformationDialog (parent, 
 			Tm_NameFromPath(path), args, num_args);
     } else
-    if (strcmp(command, MessageDialog) == 0) {
+    if (strcmp(command, xmMessageDialog) == 0) {
 	class = xmMessageBoxWidgetClass;
         XtInitializeWidgetClass(class);
         Tm_SetValues(path, interp, parent, parent, class,
@@ -729,7 +746,7 @@ Tm_DialogCmd(clientData, interp, argc, argv)
 	new = XmCreateMessageDialog (parent, 
 			Tm_NameFromPath(path), args, num_args);
     } else
-    if (strcmp(command, QuestionDialog) == 0) {
+    if (strcmp(command, xmQuestionDialog) == 0) {
 	class = xmMessageBoxWidgetClass;
         XtInitializeWidgetClass(class);
         Tm_SetValues(path, interp, parent, parent, class,
@@ -737,7 +754,7 @@ Tm_DialogCmd(clientData, interp, argc, argv)
 	new = XmCreateQuestionDialog (parent, 
 			Tm_NameFromPath(path), args, num_args);
     } else
-    if (strcmp(command, WarningDialog) == 0) {
+    if (strcmp(command, xmWarningDialog) == 0) {
 	class = xmMessageBoxWidgetClass;
         XtInitializeWidgetClass(class);
         Tm_SetValues(path, interp, parent, parent, class,
@@ -745,7 +762,7 @@ Tm_DialogCmd(clientData, interp, argc, argv)
 	new = XmCreateWarningDialog (parent, 
 			Tm_NameFromPath(path), args, num_args);
     } else
-    if (strcmp(command, WorkingDialog) == 0) {
+    if (strcmp(command, xmWorkingDialog) == 0) {
 	class = xmMessageBoxWidgetClass;
         XtInitializeWidgetClass(class);
         Tm_SetValues(path, interp, parent, parent, class,
@@ -753,7 +770,7 @@ Tm_DialogCmd(clientData, interp, argc, argv)
 	new = XmCreateWorkingDialog (parent, 
 			Tm_NameFromPath(path), args, num_args);
     } else
-    if (strcmp(command, PromptDialog) == 0) {
+    if (strcmp(command, xmPromptDialog) == 0) {
 	class = xmSelectionBoxWidgetClass;
         XtInitializeWidgetClass(class);
         Tm_SetValues(path, interp, parent, parent, class,
@@ -761,7 +778,7 @@ Tm_DialogCmd(clientData, interp, argc, argv)
 	new = XmCreatePromptDialog (parent, 
 			Tm_NameFromPath(path), args, num_args);
     } else
-    if (strcmp(command, SelectionDialog) == 0) {
+    if (strcmp(command, xmSelectionDialog) == 0) {
 	class = xmSelectionBoxWidgetClass;
         XtInitializeWidgetClass(class);
         Tm_SetValues(path, interp, parent, parent, class,
@@ -808,7 +825,7 @@ Tm_DialogCmd(clientData, interp, argc, argv)
     wPtr -> interp = interp;
     wPtr -> widget = new;
     wPtr -> pathName = XtNewString(path);
-    wPtr -> parent = dialog_path;
+    wPtr -> parent = XtNewString(dialog_path);
     wPtr -> dropProc = NULL;
     wPtr -> displayInfo = ((Tm_Widget *) parentInfo.clientData)->displayInfo;
 
@@ -821,7 +838,7 @@ Tm_DialogCmd(clientData, interp, argc, argv)
     XtAddCallback(new, XmNdestroyCallback, Tm_DestroyWidgetHandler,
 	(XtPointer) wPtr);
 
-    /* create commands for all SelectionBox children */
+    /* create commands for all xmSelectionBox children */
 
     if (class == xmSelectionBoxWidgetClass) {
 	SelectionBoxCreateChild(interp, path, new, XmDIALOG_APPLY_BUTTON, "Apply");
@@ -937,7 +954,7 @@ Tm_ScrolledCmd(clientData, interp, argc, argv)
     DivideArgs(argv + 2, argc - 2, parent_argv, &num_parent_argc,
                         child_argv, &num_child_argc);
 
-    if (strcmp(command, ScrolledList) == 0) {
+    if (strcmp(command, xmScrolledList) == 0) {
 	class = xmListWidgetClass;
         XtInitializeWidgetClass(class);
         Tm_SetValues(path, interp, parent, parent, class,
@@ -945,7 +962,7 @@ Tm_ScrolledCmd(clientData, interp, argc, argv)
         new = XmCreateScrolledList (parent,
 			Tm_NameFromPath(path), args, num_args);
     } else
-    if (strcmp(command, ScrolledText) == 0) {
+    if (strcmp(command, xmScrolledText) == 0) {
 	class = xmTextWidgetClass;
         XtInitializeWidgetClass(class);
         Tm_SetValues(path, interp, parent, parent, class,
@@ -991,7 +1008,7 @@ Tm_ScrolledCmd(clientData, interp, argc, argv)
     wPtr -> interp = interp;
     wPtr -> widget = new;
     wPtr -> pathName = XtNewString(path);
-    wPtr -> parent  = sw_path;
+    wPtr -> parent  = XtNewString(sw_path);
     wPtr -> dropProc = NULL;
     wPtr -> displayInfo = ((Tm_Widget *) parentInfo.clientData)->displayInfo;
 
@@ -1046,6 +1063,9 @@ Tm_RowColumnCmd(clientData, interp, argc, argv)
     int num_parent_argc, num_child_argc;
     Boolean managed = False;
     Tcl_CmdInfo parentInfo;
+    Tm_WidgetCmdProc widgetCmdProc;
+
+    widgetCmdProc = (Tm_WidgetCmdProc) clientData;
 
     parent = Tm_ParentWidgetFromPath (interp, path);
     if (parent == NULL)
@@ -1065,19 +1085,27 @@ Tm_RowColumnCmd(clientData, interp, argc, argv)
 
     Tm_SetValues(path, interp, parent, parent, xmRowColumnWidgetClass,
 				parent_argv, num_parent_argc, args, &num_args);
-    if (strcmp(command, RowColumn) == 0) {
+    if (strcmp(command, xmRowColumn) == 0) {
         new = XmCreateRowColumn (parent,
 			Tm_NameFromPath(path), args, num_args);
     } else 
-    if (strcmp(command, MenuBar) == 0) {
+    if (strcmp(command, xmMenuBar) == 0) {
 	new = XmCreateSimpleMenuBar(parent, 
 			Tm_NameFromPath(path), args, num_args);
     } else 
-    if (strcmp(command, PulldownMenu) == 0) {
+    if (strcmp(command, xmOptionMenu) == 0) {
+	new = XmCreateOptionMenu(parent, 
+			Tm_NameFromPath(path), args, num_args);
+    } else 
+    if (strcmp(command, xmPopupMenu) == 0) {
+	new = XmCreatePopupMenu(parent, 
+			Tm_NameFromPath(path), args, num_args);
+    } else 
+    if (strcmp(command, xmPulldownMenu) == 0) {
 	new = XmCreatePulldownMenu(parent, 
 			Tm_NameFromPath(path), args, num_args);
     } else {
-	sprintf(interp->result, "Unknown RowColumn creation command %s\n",
+	sprintf(interp->result, "Unknown xmRowColumn creation command %s\n",
 			command);
 	return TCL_ERROR;
     }
@@ -1102,7 +1130,7 @@ Tm_RowColumnCmd(clientData, interp, argc, argv)
     Tm_StoreWidgetInfo(path, wPtr, interp);
     XtVaSetValues(new, XmNuserData, wPtr, NULL);
 
-    Tcl_CreateCommand (interp, path, Tm_AnyWidgetCmd,
+    Tcl_CreateCommand (interp, path, widgetCmdProc,
 		 (ClientData) wPtr, (void (*) ()) NULL);
 
     XtAddCallback(new, XmNdestroyCallback, Tm_DestroyWidgetHandler,
