@@ -18,6 +18,185 @@
 #include "tmFuncs.h"
 #include <Xm/Xm.h>
 #include <Xm/List.h>
+#include <Xm/Text.h>
+
+/* this is a hack because some converters need info about what
+   types they are converting, and Xt does not give such info.
+   So we need to place this in globals before calling converters.
+   This is to replace the earlier, worse, Tm_MaybeSetTableSize
+ */
+char *Tm_Resource = NULL;
+
+
+/*
+ *--------------------------------------------------------------
+ *
+ * CopyToArgs --
+ *
+ *	copy to arg of correct size
+ *	borrowed from _XtCopyToArgs from Xt lib
+ *
+ * Results:
+ *
+ *	new value (of whatever type) stored in dst
+ *
+ * Side effects:
+ *
+ *--------------------------------------------------------------
+ */
+static void 
+CopyToArg(src, dst, size)
+    char* src;
+    XtArgVal *dst;
+    register unsigned int size;
+{
+    if (! (*dst)) {
+        if      (size == sizeof(long))     *dst = (XtArgVal)*(long*)src;
+        else if (size == sizeof(short))    *dst = (XtArgVal)*(short*)src;
+        else if (size == sizeof(char))     *dst = (XtArgVal)*(char*)src;
+        else if (size == sizeof(XtPointer)) *dst = (XtArgVal)*(XtPointer*)src;
+        else if (size == sizeof(char*))    *dst = (XtArgVal)*(char**)src;
+        else if (size == sizeof(XtArgVal)) *dst = *(XtArgVal*)src;
+        else bcopy((char*)src, (char*)dst, (int)size);
+    }
+} 
+
+/*
+ *--------------------------------------------------------------
+ *
+ * These functions are all to handle resources that got created
+ * and need to be reclaimed
+ *
+ *--------------------------------------------------------------
+ */
+
+static Tm_FreeResourceType *Tm_FreeResources = NULL;
+static int Tm_FreeResourcesCount;
+static int Tm_MaxFreeResourcesCount;
+
+typedef struct Tm_FreeResourceStackElmt {
+    Tm_FreeResourceType *freeResources;
+    int			freeResourcesCount;
+    int 		maxFreeResourcesCount;
+} Tm_FreeResourceStackElmt;
+
+static int Tm_FreeResourceStackDepth = 0;
+static Tm_FreeResourceStackElmt *Tm_FreeResourceStack = NULL;
+
+/*
+ *--------------------------------------------------------------
+ *
+ * Tm_InitFreeResourceList
+ *
+ *	initialise the list of resources for which space must
+ *	be reclaimed after XtSet/GetValues
+ *
+ *	to allow this to be reentrant, place old entries on stack
+ *
+ *--------------------------------------------------------------
+ */
+void Tm_InitFreeResourceList(size)
+    int size;
+{
+
+    /* store any old values if there were any */
+    if (Tm_FreeResourceStackDepth > 0) {
+	Tm_FreeResourceStack = (Tm_FreeResourceStackElmt *)
+		XtRealloc((char *)Tm_FreeResourceStack,
+			Tm_FreeResourceStackDepth * 
+			sizeof(Tm_FreeResourceStackElmt));
+	Tm_FreeResourceStack[Tm_FreeResourceStackDepth - 1].freeResources = 
+				Tm_FreeResources;
+	Tm_FreeResourceStack[Tm_FreeResourceStackDepth - 1].freeResourcesCount =
+				Tm_FreeResourcesCount;
+	Tm_FreeResourceStack[Tm_FreeResourceStackDepth - 1].maxFreeResourcesCount =
+				Tm_MaxFreeResourcesCount;
+
+    }
+    Tm_FreeResourceStackDepth++;
+
+    if (size <= 0)
+	size = 1;
+    Tm_FreeResources = (Tm_FreeResourceType *)
+			XtMalloc(size * sizeof(Tm_FreeResourceType));
+    Tm_FreeResourcesCount = 0;
+    Tm_MaxFreeResourcesCount = size;
+}
+
+
+/*
+ *--------------------------------------------------------------
+ *
+ *  Tm_AddToFreeResourceList
+ *
+ *	add some data and its corresponding free proc to the list
+ *
+ *--------------------------------------------------------------
+ */
+void  Tm_AddToFreeResourceList(data, free)
+    char *data;
+    Tm_FreeProc free;
+{
+    /* we only setup the list for adding during a Tm_Set/GetValues
+       sequence. We init at the start and free at the end. This
+       func gets called by various resource converters. However,
+       Motif (or someone else) may be calling our converter outside
+       of the Tm_Get/SetValues. We have to assume that they are doing
+       their own garbage collect, so need to make sure that we don't
+       add for their calls
+     */
+    if (Tm_FreeResources == NULL)
+	return;
+
+    /* things like XmStringTables may add many elmts. Grow if needed */
+    if (Tm_FreeResourcesCount >= Tm_MaxFreeResourcesCount) {
+	Tm_MaxFreeResourcesCount *= 2;
+	Tm_FreeResources = (Tm_FreeResourceType *) 
+				XtRealloc((char *)Tm_FreeResources,
+					Tm_MaxFreeResourcesCount * 
+					sizeof(Tm_FreeResourceType));
+    }
+    Tm_FreeResources[Tm_FreeResourcesCount].data = data;
+    Tm_FreeResources[Tm_FreeResourcesCount++].free = free;
+}
+
+
+/*
+ *--------------------------------------------------------------
+ * Tm_FreeResourceList
+ *
+ *	call the free proc on each data elmt of the list
+ *
+ *--------------------------------------------------------------
+ */
+void Tm_FreeResourceList()
+{
+    int n;
+    char *data;
+    Tm_FreeProc free;
+
+    for (n = 0; n < Tm_FreeResourcesCount; n++) {
+	free =  Tm_FreeResources[n].free;
+	data = Tm_FreeResources[n].data;
+	(free) (data);
+    }
+    XtFree((char *) Tm_FreeResources);
+    Tm_FreeResources = NULL;
+
+    Tm_FreeResourceStackDepth--;
+    if (Tm_FreeResourceStackDepth < 1)
+	return;
+
+    /* nested calls to Tm_InitFreeResourceList are unwinding - 
+       restore values from the stack
+     */
+    Tm_FreeResources =
+        Tm_FreeResourceStack[Tm_FreeResourceStackDepth - 1].freeResources;
+    Tm_FreeResourcesCount =
+        Tm_FreeResourceStack[Tm_FreeResourceStackDepth - 1].freeResourcesCount;
+    Tm_MaxFreeResourcesCount =
+        Tm_FreeResourceStack[Tm_FreeResourceStackDepth - 1].maxFreeResourcesCount;
+}
 
 /*
  *--------------------------------------------------------------
@@ -57,6 +236,9 @@ Tm_ConvertValue(w, fromType, fromValue, fromSize,
     from.addr = fromValue;
     to.size = toSize;
     to.addr = XtMalloc(toSize);
+/*
+    to.addr = NULL;
+*/
 
     if (!XtConvertAndStore(w, fromType, &from, toType, &to))  {
 	fprintf(stderr, "failed conversion %s to %s\n",
@@ -64,12 +246,10 @@ Tm_ConvertValue(w, fromType, fromValue, fromSize,
 	return FALSE;
     }
     *toValue = 0;
-    _XtCopyToArg(to.addr, toValue, to.size);
-/*
+    CopyToArg(to.addr, toValue, to.size);
     XtFree(to.addr);
-*/
 
-   return TRUE;
+    return TRUE;
 }
 
 /*
@@ -102,10 +282,12 @@ Tm_ConvertValueFromString(w, resources, num_resources,
     int n;
     XrmValue from, converted;
     XtArgVal tmpval;
+    Boolean result;
 
 #   ifdef DEBUG
     fprintf(stderr, "converting from string \"%s\" of type %s\n", orig_value, resource);
 #   endif
+
     for (n = 0; n < num_resources; n++) {
 	if (strcmp(resource, resources->resource_name) == 0) {
 	    break;
@@ -117,33 +299,13 @@ Tm_ConvertValueFromString(w, resources, num_resources,
     }
 
     /* we have a match - convert and install the resource */
-    /* special case - same type */
-    return Tm_ConvertValue(w, XtRString, orig_value,
+    Tm_Resource = resource;
+    result = Tm_ConvertValue(w, XtRString, orig_value,
 				strlen(orig_value)+1,
 				resources->resource_type,
 				new_value,
 				resources->resource_size);
-/*
-    if (strcmp(resources->resource_type, XtRString) == 0) {
-	*new_value =  orig_value;
-	return TRUE;
-    }
-    from.size = strlen(orig_value) + 1;
-    from.addr = orig_value;
-    converted.size = resources->resource_size;
-    converted.addr = XtMalloc(converted.size);
-    if (!XtConvertAndStore(w, XtRString, &from,
-		resources->resource_type, &converted)) {
-	return FALSE;
-    }
-    *new_value = 0;
-    _XtCopyToArg(converted.addr, new_value, converted.size);
-*/
-/*
-    bzero(new_value, sizeof(XtArgVal));
-    bcopy(converted.addr, new_value, converted.size);
-    *new_value = *(XtArgVal *) (converted.addr);
-*/
+    return result;
 }
 
 /*
@@ -203,13 +365,17 @@ Tm_ConvertValueFromStringQuark(w, resources, num_resources,
     from.size = strlen(orig_value) + 1;
     from.addr = orig_value;
     converted.size = resources->resource_size;
-    converted.addr = XtMalloc(converted.size);
+    converted.addr = NULL;
+
     if (!XtConvertAndStore(w, XtRString, &from,
 		XrmQuarkToString((int) resources->resource_type), &converted)) {
 	return FALSE;
     }
     *new_value = 0;
-    _XtCopyToArg(converted.addr, new_value, converted.size);
+    CopyToArg(converted.addr, new_value, converted.size);
+/* for now
+    XtFree(converted.addr);
+*/
 /*
     bzero(new_value, sizeof(XtArgVal));
     bcopy(converted.addr, new_value, converted.size);
@@ -254,6 +420,7 @@ Tm_MaybeSetStringTableSize(w, resource)
     Arg arg;
     char *size_resource;
 
+#if 0
     class = XtClass(w);
     if (class == xmListWidgetClass) {
 	if (strcmp (resource, XmNitems) == 0) {
@@ -269,6 +436,7 @@ Tm_MaybeSetStringTableSize(w, resource)
     /* Sigh, we have a StringTable, find its size */
     XtSetArg(arg, size_resource, &Tm_XmStringTableSize);
     XtGetValues(w, &arg, 1);
+#endif
 }
 
 /*
@@ -304,12 +472,24 @@ Tm_ConvertResourceToString(w, resource, resource_size, resource_type, value)
     
     /* special case - same type */
     if (strcmp(resource_type, XtRString) == 0) {
-	*value =  *(char **) from.addr;
+	*value =  XtNewString(*(char **) from.addr);
+	/* check if it is Text's value */
+	if (XtIsSubclass(w, xmTextWidgetClass) &&
+	    strcmp(resource, XmNvalue) == 0)  {
+	    XtFree(*(char **) from.addr);
+	}
+	XtFree(from.addr);
+
+	/* allow this to be gc later */
+	Tm_AddToFreeResourceList(*value, (Tm_FreeProc) XtFree);
+
 	return TRUE;
     }
+
     converted.addr = NULL;
     if (!XtConvertAndStore(w, resource_type, &from,
 			XtRString, &converted)) {
+	XtFree(from.addr);
 	return FALSE;
     }
     bzero(value, sizeof(XtArgVal));
@@ -354,6 +534,7 @@ Tm_ConvertValueToString(w, resources, num_resources,
 	return FALSE;
     }
 
+    Tm_Resource = resource;
     return Tm_ConvertResourceToString(w, resource,
 		resources->resource_size, resources->resource_type, new_value);
 
@@ -530,19 +711,28 @@ Tm_SetValues(pathName, interp, w, parent, class, argv, argc, args, num_args)
 	resource = argv[0]+1;
         value = argv[1];
 
+        /* we cant allow callbacks to be converted - too messy */
+        if (strstr(resource, "Callback") != NULL) {
+	    sprintf(interp->result, "setValues doesn't support callbacks.\
+Use \"%s\" method\n", resource);
+  	    return TCL_ERROR;
+        }
+
         if (Tm_ConvertValueFromString(w, resources, num_resources,
 		resource, value, &new_value)) {
             XtSetArg(args[num_values], resource, new_value);
 	    num_values++;
 	} else
-        if (Tm_ConvertValueFromString(w, constraint_resources, 
+        if (num_constraint_resources > 0 &&
+	    Tm_ConvertValueFromString(w, constraint_resources, 
 	            num_constraint_resources, resource,
         	    value, &new_value)) {
             XtSetArg(args[num_values], resource, new_value);
 	    num_values++;
 	} else 
 #ifndef MOTIF11
-        if (Tm_ConvertSubValueFromString(w, sec_resources, 
+        if (num_sec_resources > 0 &&
+	    Tm_ConvertSubValueFromString(w, sec_resources, 
 	            num_sec_resources, resource,
         	    value, &new_value)) {
             XtSetArg(args[num_values], resource, new_value);
@@ -558,13 +748,17 @@ Tm_SetValues(pathName, interp, w, parent, class, argv, argc, args, num_args)
 	argv += 2;
     }
 
+#ifndef MOTIF11
+    if (num_sec_resources > 0)
+	Tm_FreeSecondaryResources(sec_resources, num_sec_resources);
+#endif /* MOTIF11 */
+    if (num_resources > 0) 
+	XtFree((char *) resources);
+    if (num_constraint_resources > 0) 
+	XtFree((char *) constraint_resources);
+
     *num_args = num_values;
     return TCL_OK;
-
-/*
-    Tm_FreeSecondaryResources(sec_resources, num_sec_resources);
-*/
-    /* memory leak: should XtFree all new_values stored in args[] */
 }
 
 /*
@@ -641,7 +835,8 @@ Tm_GetValues(pathName, interp, w, class, argv, argc)
 			TCL_LEAVE_ERR_MSG) == NULL)
 		fprintf(stderr, "%s\n", interp->result);
 	} else
-        if (Tm_ConvertValueToString(w, constraint_resources, 
+        if (num_constraint_resources > 0 &&
+	    Tm_ConvertValueToString(w, constraint_resources, 
 	            num_constraint_resources, resource,
         	    &new_value)) {
 	    /* store in a variable now */
@@ -650,7 +845,8 @@ Tm_GetValues(pathName, interp, w, class, argv, argc)
 		fprintf(stderr, "%s\n", interp->result);
 	} else 
 #ifndef MOTIF11
-        if (Tm_ConvertSubValueToString(w, sec_resources, 
+        if (num_sec_resources > 0 &&
+	    Tm_ConvertSubValueToString(w, sec_resources, 
 	            num_sec_resources, resource,
         	    &new_value)) {
 	    /* store in a variable now */
@@ -668,8 +864,13 @@ Tm_GetValues(pathName, interp, w, class, argv, argc)
 	argv += 2;
     }
 #ifndef MOTIF11
-    Tm_FreeSecondaryResources(sec_resources, num_sec_resources);
+    if (num_sec_resources > 0)
+	Tm_FreeSecondaryResources(sec_resources, num_sec_resources);
 #endif
+    if (num_resources > 0) 
+	XtFree((char *) resources);
+    if (num_constraint_resources > 0) 
+	XtFree((char *) constraint_resources);
     return TCL_OK;
 }
 
@@ -702,6 +903,9 @@ Tm_GetResourceInfo(resourceInfo, w, resource)
     Tcl_DStringAppendElement(resourceInfo, resource->resource_class);
     Tcl_DStringAppendElement(resourceInfo, resource->resource_type);
 
+    /* hack to let resource converter know what this is */
+    Tm_Resource = resource->resource_name;
+
     if (Tm_ConvertResourceToString(w, resource->resource_name,
 		resource->resource_size, resource->resource_type,
 		&resource_value)) {;
@@ -792,14 +996,24 @@ Tm_ResourceList(interp, w, class)
     oldWarningHandler = XtAppSetWarningHandler(XtWidgetToApplicationContext(w), 
 					Tm_NoErrorMsg);
 
+    /* set up for garbage collect of resource values */
+    Tm_InitFreeResourceList(128);
+
     Tcl_DStringInit(&resourceList);
 
     Tcl_DStringInit(&resourceInfo);
     for (n = 0; n < num_resources; n++) {
 	Tm_GetResourceInfo(&resourceInfo, w, resources+n);
 	/* hack to get string out of D-string */
+/*
 	Tcl_DStringResult(interp, &resourceInfo);
 	Tcl_DStringAppendElement(&resourceList, interp->result);
+*/
+
+	/* no, do it the proper way (accessing internals of DString :-(  ) */
+	Tcl_DStringAppendElement(&resourceList, resourceInfo.string);
+	Tcl_DStringFree(&resourceInfo);
+	Tcl_DStringInit(&resourceInfo);
     }
 
 #   ifndef MOTIF11
@@ -826,6 +1040,17 @@ Tm_ResourceList(interp, w, class)
     XtAppSetWarningHandler(XtWidgetToApplicationContext(w), oldWarningHandler);
 
     Tcl_DStringResult(interp, &resourceList);
+
+#ifndef MOTIF11
+    if (num_sec_resources > 0)
+	Tm_FreeSecondaryResources(sec_resources, num_sec_resources);
+#endif
+    if (num_resources > 0) 
+	XtFree((char *) resources);
+    if (num_constraint_resources > 0) 
+	XtFree((char *) constraint_resources);
+    Tm_FreeResourceList();
+
     return TCL_OK;
 }
 
